import numpy as np
from astrowidgets import ImageWidget
from photutils import centroid_com
import ipywidgets as ipw
from astropy.stats import sigma_clipped_stats
from astropy.table import Table
import matplotlib.pyplot as plt
from glowing_waffles.visualization import seeing_plot
out = ipw.Output()
out2 = ipw.Output()
out3 = ipw.Output()

def build(imagewidget):

    bind_map = imagewidget._viewer.get_bindmap()
    # Displays the event map...
    #bind_map.eventmap
    bind_map.clear_event_map()
    bind_map.map_event(None, (), 'ms_left', 'pan')
    bind_map.map_event(None, (), 'pa_pan', 'zoom')

    #bind_map.map_event(None, (), 'ms_left', 'cursor')
    # contrast with right mouse
    bind_map.map_event(None, (), 'ms_right', 'contrast')

    # shift-right mouse to reset contrast
    bind_map.map_event(None, ('shift',), 'ms_right', 'contrast_restore')
    bind_map.map_event(None, ('ctrl',), 'ms_left', 'cursor')

    # Bind +/- to zoom in/out
    bind_map.map_event(None, (), 'kp_+', 'zoom_in')
    bind_map.map_event(None, (), 'kp_=', 'zoom_in')
    bind_map.map_event(None, (), 'kp_-', 'zoom_out')
    bind_map.map_event(None, (), 'kp__', 'zoom_out')

    # Bind arrow keys to panning
    # There is NOT a typo below. I want the keys to move the image in the
    # direction of the arrow
    bind_map.map_event(None, (), 'kp_left', 'pan_right')
    bind_map.map_event(None, (), 'kp_right', 'pan_left')
    bind_map.map_event(None, (), 'kp_up', 'pan_down')
    bind_map.map_event(None, (), 'kp_down', 'pan_up')

def make_show_event(iw):
    def show_event(viewer, event, datax, datay):

        i = iw._viewer.get_image()
        data = i.get_data()
        pad = 15
        x = int(np.floor(event.data_x))
        y = int(np.floor(event.data_y))
        cnt = 0
        sub_data = data[y - pad:y + pad, x - pad:x + pad] #- med
        _, sub_med, _ = sigma_clipped_stats(sub_data)
        #sub_med = 0
        foo, moo = centroid_com(sub_data - sub_med)
        cenx = foo + x - pad
        ceny = moo + y - pad
        cen = np.array([foo + x - pad, moo + y - pad])
        ceno = np.array([-100, -100])
        while cnt <= 10 and (np.abs(np.array([foo, moo]) - pad).max() >3 or np.abs(cen - ceno).max() > 0.1):
           # print(cnt, foo, moo)
            x = int(np.floor(foo)) + x - pad
            y = int(np.floor(moo)) + y - pad
            sub_data = data[y - pad:y + pad, x - pad:x + pad] #- med
            _, sub_med, _ = sigma_clipped_stats(sub_data)
            #sub_med = 0
            mask = (sub_data - sub_med) < 0
            foo, moo = centroid_com(sub_data - sub_med, mask=mask)
            ceno = cen
            cen = np.array([foo + x - pad, moo + y - pad])
    #             print(cen)
    #             print(cen - ceno)
            cnt += 1

        iw.add_markers(Table(data=[[cen[0]], [cen[1]]], names=['x', 'y']))
        #print(foo, moo)
        yd, xd = np.indices((sub_data.shape))
        r = np.sqrt((xd - foo)**2 + (yd - moo)**2)
        r_exact = r.copy()
        r = r.astype(np.int)
        tbin = np.bincount(r.ravel(), sub_data.ravel())
        rbin = np.bincount(r.ravel(), r_exact.ravel())
        nr = np.bincount(r.ravel())
        radialprofile = tbin / nr
        ravg = rbin / nr
        adjust_max = radialprofile.max() - sub_med
        scaled_profile = (radialprofile - sub_med) / adjust_max
        scaled_exact_counts = (sub_data - sub_med) / adjust_max
        out.clear_output(wait=True)
        with out:
           # print(dir(event))
            #print(event.data_x, event.data_y)
            plt.clf()
            #sub_med += med
            seeing_plot(r_exact, scaled_exact_counts, ravg, scaled_profile, 4.5,
                    'Some Image Name', file_name='some_name', gap=6, annulus_width=13)
            plt.show()
        out2.clear_output(wait=True)
        with out2:
            tbin2 = np.bincount(r.ravel(), (sub_data - sub_med).ravel())
            counts = np.cumsum(tbin2)
            mag_diff = -2.5 * np.log10(counts/counts.max())
            plt.plot(range(len(radialprofile)), counts)
            plt.xlim(0, 20)
            #plt.ylim(0.2, 0)
            plt.grid()
            sub_blot = sub_data.copy()
            sub_blot[10:30, 10:30] = np.nan
            sub_std = np.nanstd(sub_blot)
            plt.title('Net counts in aperture std {:.2f} med {:.2f}'.format(sub_std, sub_med))
            sub_pois = (sub_data - sub_med)
            e_sky = np.sqrt(sub_med)
            rn = 10
            sub_noise_sq = np.sqrt(sub_pois ** 2 + sub_std ** 2) ** 2
            nbin = np.sqrt(np.bincount(r.ravel(), (sub_noise_sq).ravel()))
            plt.xlabel('Aperture radius')
            plt.show()
        out3.clear_output(wait=True)
        with out3:
            poisson = np.sqrt(np.cumsum(tbin2))
            error = np.sqrt(poisson ** 2 + np.cumsum(nr) * (e_sky ** 2 + rn ** 2))
            snr = np.cumsum(tbin2) / error
            snr_max = snr[:20].max()
            plt.plot(range(len(radialprofile)), snr)
            plt.title('Signal to noise ratio {}'.format(snr.max()))
            plt.xlim(0, 20)
            #plt.ylim(0, 2)
            plt.xlabel('Aperture radius')
            plt.grid()
            plt.show()
    return show_event

def box(imagewidget):
    big_box = ipw.HBox()
    layout = ipw.Layout(width='20ch')
    hb = ipw.HBox()
    ap_t = ipw.IntText(description='Aperture', value=5, layout=layout)
    in_t = ipw.IntText(description='Inner annulus', value=10, layout=layout)
    out_t = ipw.IntText(description='Outer annulus', value=20, layout=layout)
    hb.children = [ap_t, in_t, out_t]

    lil_box = ipw.VBox()
    lil_box.children = [out, out2, out3]
    big_box.children = [imagewidget, lil_box]
    big_box.layout.width = '100%'
    # Line below puts space between the image and the plots so the plots
    # don't jump around as the image value changes.
    big_box.layout.justify_content = 'space-between'

    return big_box




"""Module containing core functionality of ``astrowidgets``."""

# STDLIB
import functools
import warnings

# THIRD-PARTY
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table, vstack

# Jupyter widgets
import ipywidgets as ipyw

# Ginga
from ginga.AstroImage import AstroImage
from ginga.canvas.CanvasObject import drawCatalog
from ginga.web.jupyterw.ImageViewJpw import EnhancedCanvasView
from ginga.util.wcs import raDegToString, decDegToString

__all__ = ['ImageWidget']

# Allowed locations for cursor display
ALLOWED_CURSOR_LOCATIONS = ['top', 'bottom', None]

# List of marker names that are for internal use only
RESERVED_MARKER_SET_NAMES = ['all']


class ImageWidget(ipyw.VBox):
    """
    Image widget for Jupyter notebook using Ginga viewer.

    .. todo:: Any property passed to constructor has to be valid keyword.

    Parameters
    ----------
    logger : obj or ``None``
        Ginga logger. For example::

            from ginga.misc.log import get_logger
            logger = get_logger('my_viewer', log_stderr=False,
                                log_file='ginga.log', level=40)

    image_width, image_height : int
        Dimension of Jupyter notebook's image widget.

    use_opencv : bool
        Let Ginga use ``opencv`` to speed up image transformation;
        e.g., rotation and mosaic. If this is enabled and you
        do not have ``opencv``, you will get a warning.

    pixel_coords_offset : int, optional
        An offset, typically either 0 or 1, to add/subtract to all
        pixel values when going to/from the displayed image.
        *In almost all situations the default value, ``0``, is the
        correct value to use.*

    """

    def __init__(self, logger=None, image_width=500, image_height=500,
                 use_opencv=True, pixel_coords_offset=0):
        super().__init__()

        # TODO: Is this the best place for this?
        if use_opencv:
            try:
                from ginga import trcalc
                trcalc.use('opencv')
            except ImportError:
                warnings.warn('install opencv or set use_opencv=False')

        self._viewer = EnhancedCanvasView(logger=logger)

        self._pixel_offset = pixel_coords_offset

        self._jup_img = ipyw.Image(format='jpeg')

        # Set the image margin to over the widgets default of 2px on
        # all sides.
        self._jup_img.layout.margin = '0'

        # Set both of those to ensure consistent display in notebook
        # and jupyterlab when the image is put into a container smaller
        # than the image.

        self._jup_img.max_width = '100%'
        self._jup_img.height = 'auto'

        # Set the width of the box containing the image to the desired width
        self.layout.width = str(image_width)

        # Note we are NOT setting the height. That is because the height
        # is automatically set by the image aspect ratio.

        # These need to also be set for now; ginga uses them to figure
        # out what size image to make.
        self._jup_img.width = image_width
        self._jup_img.height = image_height

        self._viewer.set_widget(self._jup_img)

        # enable all possible keyboard and pointer operations
        self._viewer.get_bindings().enable_all(True)

        # enable draw
        self.dc = drawCatalog
        self.canvas = self.dc.DrawingCanvas()
        self.canvas.enable_draw(True)
        self.canvas.enable_edit(True)

        # Make sure all of the internal state trackers have a value
        # and start in a state which is definitely allowed: all are
        # False.
        self._is_marking = False
        self._click_center = False
        self._click_drag = False
        self._scroll_pan = False

        # Set a couple of things to match the ginga defaults
        self.scroll_pan = True
        self.click_drag = False

        bind_map = self._viewer.get_bindmap()
        # Set up right-click and drag adjusts the contrast
        bind_map.map_event(None, (), 'ms_right', 'contrast')
        # Shift-right-click restores the default contrast
        bind_map.map_event(None, ('shift',), 'ms_right', 'contrast_restore')

        # Marker
        self.marker = {'type': 'circle', 'color': 'cyan', 'radius': 20}
        # Maintain marker tags as a set because we do not want
        # duplicate names.
        self._marktags = set()
        # Let's have a default name for the tag too:
        self._default_mark_tag_name = 'default-marker-name'
        self._interactive_marker_set_name_default = 'interactive-markers'
        self._interactive_marker_set_name = self._interactive_marker_set_name_default

        # coordinates display
        self._jup_coord = ipyw.HTML('Coordinates show up here')
        # This needs ipyevents 0.3.1 to work
        self._viewer.add_callback('cursor-changed', self._mouse_move_cb)
        self._viewer.add_callback('cursor-down', self._mouse_click_cb)

        # Define a callback that shows the output of a print
        self.print_out = ipyw.Output()

        self._cursor = 'bottom'
        self.children = [self._jup_img, self._jup_coord]

    @property
    def logger(self):
        """Logger for this widget."""
        return self._viewer.logger

    @property
    def image_width(self):
        return int(self._jup_img.width)

    @image_width.setter
    def image_width(self, value):
        # widgets expect width/height as strings, but most users will not, so
        # do the conversion.
        self._jup_img.width = str(value)
        self._viewer.set_window_size(self.image_width, self.image_height)

    @property
    def image_height(self):
        return int(self._jup_img.height)

    @image_height.setter
    def image_height(self, value):
        # widgets expect width/height as strings, but most users will not, so
        # do the conversion.
        self._jup_img.height = str(value)
        self._viewer.set_window_size(self.image_width, self.image_height)

    @property
    def pixel_offset(self):
        """
        An offset, typically either 0 or 1, to add/subtract to all
        pixel values when going to/from the displayed image.
        *In almost all situations the default value, ``0``, is the
        correct value to use.*

        This value cannot be modified after initialization.
        """
        return self._pixel_offset

    def _mouse_move_cb(self, viewer, button, data_x, data_y):
        """
        Callback to display position in RA/DEC deg.
        """
        if self.cursor is None:  # no-op
            return

        image = viewer.get_image()
        if image is not None:
            ix = int(data_x + 0.5)
            iy = int(data_y + 0.5)
            try:
                imval = viewer.get_data(ix, iy)
                imval = '{:8.3f}'.format(imval)
            except Exception:
                imval = 'N/A'

            val = 'X: {:.2f}, Y: {:.2f}'.format(data_x + self._pixel_offset,
                                                data_y + self._pixel_offset)
            if image.wcs.wcs is not None:
                ra, dec = image.pixtoradec(data_x, data_y)
                val += ' (RA: {}, DEC: {})'.format(
                    raDegToString(ra), decDegToString(dec))
            val += ', value: {}'.format(imval)
            self._jup_coord.value = val

    def _mouse_click_cb(self, viewer, event, data_x, data_y):
        """
        Callback to handle mouse clicks.
        """
        if self.is_marking:
            marker_name = self._interactive_marker_set_name
            objs = []
            try:
                c_mark = viewer.canvas.get_object_by_tag(marker_name)
            except Exception:  # Nothing drawn yet
                pass
            else:  # Add to existing marks
                objs = c_mark.objects
                viewer.canvas.delete_object_by_tag(marker_name)

            # NOTE: By always using CompoundObject, marker handling logic
            # is simplified.
            obj = self._marker(x=data_x, y=data_y)
            objs.append(obj)
            viewer.canvas.add(self.dc.CompoundObject(*objs),
                              tag=marker_name)
            self._marktags.add(marker_name)
            with self.print_out:
                print('Selected {} {}'.format(obj.x, obj.y))

        elif self.click_center:
            self.center_on((data_x, data_y))

            with self.print_out:
                print('Centered on X={} Y={}'.format(data_x + self._pixel_offset,
                                                     data_y + self._pixel_offset))

#     def _repr_html_(self):
#         """
#         Show widget in Jupyter notebook.
#         """
#         from IPython.display import display
#         return display(self._widget)

    def load_fits(self, fitsorfn, numhdu=None, memmap=None):
        """
        Load a FITS file into the viewer.

        Parameters
        ----------
        fitsorfn : str or HDU
            Either a file name or an HDU (*not* an HDUList).
            If file name is given, WCS in primary header is automatically
            inherited. If a single HDU is given, WCS must be in the HDU
            header.

        numhdu : int or ``None``
            Extension number of the desired HDU.
            If ``None``, it is determined automatically.

        memmap : bool or ``None``
            Memory mapping.
            If ``None``, it is determined automatically.

        """
        if isinstance(fitsorfn, str):
            image = AstroImage(logger=self.logger, inherit_primary_header=True)
            image.load_file(fitsorfn, numhdu=numhdu, memmap=memmap)
            self._viewer.set_image(image)

        elif isinstance(fitsorfn, (fits.ImageHDU, fits.CompImageHDU,
                                   fits.PrimaryHDU)):
            self._viewer.load_hdu(fitsorfn)

    def load_nddata(self, nddata):
        """
        Load an ``NDData`` object into the viewer.

        .. todo:: Add flag/masking support, etc.

        Parameters
        ----------
        nddata : `~astropy.nddata.NDData`
            ``NDData`` with image data and WCS.

        """
        from ginga.util.wcsmod.wcs_astropy import AstropyWCS

        image = AstroImage(logger=self.logger)
        image.set_data(nddata.data)
        _wcs = AstropyWCS(self.logger)
        if nddata.wcs:
            _wcs.load_header(nddata.wcs.to_header())

        try:
            image.set_wcs(_wcs)
        except Exception as e:
            print('Unable to set WCS from NDData: {}'.format(str(e)))
        self._viewer.set_image(image)

    def load_array(self, arr):
        """
        Load a 2D array into the viewer.

        .. note:: Use :meth:`load_nddata` for WCS support.

        Parameters
        ----------
        arr : array-like
            2D array.

        """
        self._viewer.load_data(arr)

    def center_on(self, point):
        """
        Centers the view on a particular point.

        Parameters
        ----------
        point : tuple or `~astropy.coordinates.SkyCoord`
            If tuple of ``(X, Y)`` is given, it is assumed
            to be in data coordinates.
        """
        if isinstance(point, SkyCoord):
            self._viewer.set_pan(point.ra.deg, point.dec.deg, coord='wcs')
        else:
            self._viewer.set_pan(*(np.asarray(point) - self._pixel_offset))

    def offset_to(self, dx, dy, skycoord_offset=False):
        """
        Move the center to a point that is given offset
        away from the current center.

        Parameters
        ----------
        dx, dy : float
            Offset value. Unit is assumed based on
            ``skycoord_offset``.

        skycoord_offset : bool
            If `True`, offset must be given in degrees.
            Otherwise, they are in pixel values.

        """
        if skycoord_offset:
            coord = 'wcs'
        else:
            coord = 'data'

        pan_x, pan_y = self._viewer.get_pan(coord=coord)
        self._viewer.set_pan(pan_x + dx, pan_y + dy, coord=coord)

    @property
    def zoom_level(self):
        """
        Zoom level:

        * 1 means real-pixel-size.
        * 2 means zoomed in by a factor of 2.
        * 0.5 means zoomed out by a factor of 2.

        """
        return self._viewer.get_scale()

    @zoom_level.setter
    def zoom_level(self, val):
        if val == 'fit':
            self._viewer.zoom_fit()
        else:
            self._viewer.scale_to(val, val)

    def zoom(self, val):
        """
        Zoom in or out by the given factor.

        Parameters
        ----------
        val : int
            The zoom level to zoom the image.
            See `zoom_level`.

        """
        self.zoom_level = self.zoom_level * val

    @property
    def is_marking(self):
        """
        `True` if in marking mode, `False` otherwise.
        Marking mode means a mouse click adds a new marker.
        This does not affect :meth:`add_markers`.
        """
        return self._is_marking

    def start_marking(self, marker_name=None,
                      marker=None):
        """
        Start marking, with option to name this set of markers or
        to specify the marker style.
        """
        self._cached_state = dict(click_center=self.click_center,
                                  click_drag=self.click_drag,
                                  scroll_pan=self.scroll_pan)
        self.click_center = False
        self.click_drag = False
        # Set scroll_pan to ensure there is a mouse way to pan
        self.scroll_pan = True
        self._is_marking = True
        if marker_name is not None:
            self._validate_marker_name(marker_name)
            self._interactive_marker_set_name = marker_name
            self._marktags.add(marker_name)
        else:
            self._interactive_marker_set_name = \
                self._interactive_marker_set_name_default
        if marker is not None:
            self.marker = marker

    def stop_marking(self, clear_markers=False):
        """
        Stop marking mode, with option to clear markers, if desired.

        Parameters
        ----------
        clear_markers : bool, optional
            If ``clear_markers`` is `False`, existing markers are
            retained until :meth:`reset_markers` is called.
            Otherwise, they are erased.
        """
        if self.is_marking:
            self._is_marking = False
            self.click_center = self._cached_state['click_center']
            self.click_drag = self._cached_state['click_drag']
            self.scroll_pan = self._cached_state['scroll_pan']
            self._cached_state = {}
            if clear_markers:
                self.reset_markers()

    @property
    def marker(self):
        """
        Marker to use.

        .. todo:: Add more examples.

        Marker can be set as follows::

            {'type': 'circle', 'color': 'cyan', 'radius': 20}
            {'type': 'cross', 'color': 'green', 'radius': 20}
            {'type': 'plus', 'color': 'red', 'radius': 20}

        """
        # Change the marker from a very ginga-specific type (a partial
        # of a ginga drawing canvas type) to a generic dict, which is
        # what we expect the user to provide.
        #
        # That makes things like self.marker = self.marker work.
        return self._marker_dict

    @marker.setter
    def marker(self, val):
        # Make a new copy to avoid modifying the dict that the user passed in.
        _marker = val.copy()
        marker_type = _marker.pop('type')
        if marker_type == 'circle':
            self._marker = functools.partial(self.dc.Circle, **_marker)
        elif marker_type == 'plus':
            _marker['type'] = 'point'
            _marker['style'] = 'plus'
            self._marker = functools.partial(self.dc.Point, **_marker)
        elif marker_type == 'cross':
            _marker['type'] = 'point'
            _marker['style'] = 'cross'
            self._marker = functools.partial(self.dc.Point, **_marker)
        else:  # TODO: Implement more shapes
            raise NotImplementedError(
                'Marker type "{}" not supported'.format(marker_type))
        # Only set this once we have successfully created a marker
        self._marker_dict = val

    def get_markers(self, x_colname='x', y_colname='y',
                    skycoord_colname='coord',
                    marker_name=None):
        """
        Return the locations of existing markers.

        Parameters
        ----------
        x_colname, y_colname : str
            Column names for X and Y data coordinates.
            Coordinates returned are 0- or 1-indexed, depending
            on ``self.pixel_offset``.

        skycoord_colname : str
            Column name for ``SkyCoord``, which contains
            sky coordinates associated with the active image.
            This is ignored if image has no WCS.

        Returns
        -------
        markers_table : `~astropy.table.Table` or ``None``
            Table of markers, if any, or ``None``.

        """
        if marker_name is None:
            marker_name = self._default_mark_tag_name

        if marker_name == 'all':
            # If it wasn't for the fact that SKyCoord columns can't
            # be stacked this would all fit nicely into a list
            # comprehension. But they can't, so we delete the
            # SkyCoord column if it is present, then add it
            # back after we have stacked.
            coordinates = []
            tables = []
            for name in self._marktags:
                table = self.get_markers(x_colname=x_colname,
                                         y_colname=y_colname,
                                         skycoord_colname=skycoord_colname,
                                         marker_name=name)
                if table is None:
                    # No markers by this name, skip it
                    continue

                try:
                    coordinates.extend(c for c in table[skycoord_colname])
                except KeyError:
                    pass
                else:
                    del table[skycoord_colname]
                tables.append(table)

            stacked = vstack(tables, join_type='exact')

            if coordinates:
                stacked[skycoord_colname] = SkyCoord(coordinates)

            return stacked

        # We should always allow the default name. The case
        # where that table is empty will be handled in a moment.
        if (marker_name not in self._marktags
                and marker_name != self._default_mark_tag_name):
            raise ValueError(f"No markers named '{marker_name}' found.")

        try:
            c_mark = self._viewer.canvas.get_object_by_tag(marker_name)
        except Exception:
            # No markers in this table. Issue a warning and continue
            warnings.warn(f"Marker set named '{marker_name}' is empty",
                          category=UserWarning)
            return None

        image = self._viewer.get_image()
        xy_col = []

        if (image is None) or (image.wcs.wcs is None):
            # Do not include SkyCoord column
            include_skycoord = False
        else:
            include_skycoord = True
            radec_col = []

        # Extract coordinates from markers
        for obj in c_mark.objects:
            if obj.coord == 'data':
                xy_col.append([obj.x, obj.y])
                if include_skycoord:
                    radec_col.append([np.nan, np.nan])
            elif not include_skycoord:  # marker in WCS but image has none
                self.logger.warning(
                    'Skipping ({},{}); image has no WCS'.format(obj.x, obj.y))
            else:  # wcs
                xy_col.append([np.nan, np.nan])
                radec_col.append([obj.x, obj.y])

        # Convert to numpy arrays
        xy_col = np.asarray(xy_col)  # [[x0, y0], [x1, y1], ...]

        if include_skycoord:
            # [[ra0, dec0], [ra1, dec1], ...]
            radec_col = np.asarray(radec_col)

            # Fill in X,Y from RA,DEC
            mask = np.isnan(xy_col[:, 0])  # One bool per row
            if np.any(mask):
                xy_col[mask] = image.wcs.wcspt_to_datapt(radec_col[mask])

            # Fill in RA,DEC from X,Y
            mask = np.isnan(radec_col[:, 0])
            if np.any(mask):
                radec_col[mask] = image.wcs.datapt_to_wcspt(xy_col[mask])

            sky_col = SkyCoord(radec_col[:, 0], radec_col[:, 1], unit='deg')

        # Convert X,Y from 0-indexed to 1-indexed
        if self._pixel_offset != 0:
            xy_col += self._pixel_offset

        # Build table
        if include_skycoord:
            markers_table = Table(
                [xy_col[:, 0], xy_col[:, 1], sky_col],
                names=(x_colname, y_colname, skycoord_colname))
        else:
            markers_table = Table(xy_col, names=(x_colname, y_colname))

        # Either way, add the marker names
        markers_table['marker name'] = marker_name
        return markers_table

    def _validate_marker_name(self, marker_name):
        """
        Raise an error if the marker_name is not allowed.
        """
        if marker_name in RESERVED_MARKER_SET_NAMES:
            raise ValueError('The marker name {} is not allowed. Any name is '
                             'allowed except these: '
                             '{}'.format(marker_name,
                                         ', '.join(RESERVED_MARKER_SET_NAMES)))

    def add_markers(self, table, x_colname='x', y_colname='y',
                    skycoord_colname='coord', use_skycoord=False,
                    marker_name=None):
        """
        Creates markers in the image at given points.

        .. todo::

            Later enhancements to include more columns
            to control size/style/color of marks,

        Parameters
        ----------
        table : `~astropy.table.Table`
            Table containing marker locations.

        x_colname, y_colname : str
            Column names for X and Y.
            Coordinates can be 0- or 1-indexed, as
            given by ``self.pixel_offset``.

        skycoord_colname : str
            Column name with ``SkyCoord`` objects.

        use_skycoord : bool
            If `True`, use ``skycoord_colname`` to mark.
            Otherwise, use ``x_colname`` and ``y_colname``.

        marker_name : str, optional
            Name to assign the markers in the table. Providing a name
            allows markers to be removed by name at a later time.
        """
        # TODO: Resolve https://github.com/ejeschke/ginga/issues/672

        # For now we always convert marker locations to pixels; see
        # comment below.
        coord_type = 'data'

        if marker_name is None:
            marker_name = self._default_mark_tag_name

        self._validate_marker_name(marker_name)

        self._marktags.add(marker_name)

        # Extract coordinates from table.
        # They are always arrays, not scalar.
        if use_skycoord:
            image = self._viewer.get_image()
            if image is None:
                raise ValueError('Cannot get image from viewer')
            if image.wcs.wcs is None:
                raise ValueError(
                    'Image has no valid WCS, '
                    'try again with use_skycoord=False')
            coord_val = table[skycoord_colname]
            # TODO: Maybe switch back to letting ginga handle conversion
            #       to pixel coordinates.
            # Convert to pixels here (instead of in ginga) because conversion
            # in ginga is currently very slow.
            coord_x, coord_y = image.wcs.wcs.all_world2pix(coord_val.ra.deg,
                                                           coord_val.dec.deg,
                                                           0)
            # In the event a *single* marker has been added, coord_x and coord_y
            # will be scalars. Make them arrays always.
            if np.ndim(coord_x) == 0:
                coord_x = np.array([coord_x])
                coord_y = np.array([coord_y])
        else:  # Use X,Y
            coord_x = table[x_colname].data
            coord_y = table[y_colname].data
            # Convert data coordinates from 1-indexed to 0-indexed
            if self._pixel_offset != 0:
                # Don't use the in-place operator -= here...that modifies
                # the input table.
                coord_x = coord_x - self._pixel_offset
                coord_y = coord_y - self._pixel_offset

        # Prepare canvas and retain existing marks
        objs = []
        try:
            c_mark = self._viewer.canvas.get_object_by_tag(marker_name)
        except Exception:
            pass
        else:
            objs = c_mark.objects
            self._viewer.canvas.delete_object_by_tag(marker_name)

        # TODO: Test to see if we can mix WCS and data on the same canvas
        objs += [self._marker(x=x, y=y, coord=coord_type)
                 for x, y in zip(coord_x, coord_y)]
        self._viewer.canvas.add(self.dc.CompoundObject(*objs),
                                tag=marker_name)

    def remove_markers(self, marker_name=None):
        """
        Remove some but not all of the markers by name used when
        adding the markers

        Parameters
        ----------

        marker_name : str, optional
            Name used when the markers were added.
        """
        # TODO:
        #   arr : ``SkyCoord`` or array-like
        #   Sky coordinates or 2xN array.
        #
        # NOTE: How to match? Use np.isclose?
        #       What if there are 1-to-many matches?

        if marker_name is None:
            marker_name = self._default_mark_tag_name

        if marker_name not in self._marktags:
            # This shouldn't have happened, raise an error
            raise ValueError('Marker name {} not found in current markers.'
                             ' Markers currently in use are '
                             '{}'.format(marker_name,
                                         sorted(self._marktags)))

        try:
            self._viewer.canvas.delete_object_by_tag(marker_name)
        except KeyError:
            raise KeyError('Unable to remove markers named {} from image. '
                           ''.format(marker_name))
        else:
            self._marktags.remove(marker_name)

    def reset_markers(self):
        """
        Delete all markers.
        """

        # Grab the entire list of marker names before iterating
        # otherwise what we are iterating over changes.
        for marker_name in list(self._marktags):
            self.remove_markers(marker_name)

    @property
    def stretch_options(self):
        """
        List all available options for image stretching.
        """
        return self._viewer.get_color_algorithms()

    @property
    def stretch(self):
        """
        The image stretching algorithm in use.
        """
        return self._viewer.rgbmap.dist

    # TODO: Possible to use astropy.visualization directly?
    @stretch.setter
    def stretch(self, val):
        valid_vals = self.stretch_options
        if val not in valid_vals:
            raise ValueError('Value must be one of: {}'.format(valid_vals))
        self._viewer.set_color_algorithm(val)

    @property
    def autocut_options(self):
        """
        List all available options for image auto-cut.
        """
        return self._viewer.get_autocut_methods()

    @property
    def cuts(self):
        """
        Current image cut levels.
        To set new cut levels, either provide a tuple of
        ``(low, high)`` values or one of the options from
        `autocut_options`.
        """
        return self._viewer.get_cut_levels()

    # TODO: Possible to use astropy.visualization directly?
    @cuts.setter
    def cuts(self, val):
        if isinstance(val, str):  # Autocut
            valid_vals = self.autocut_options
            if val not in valid_vals:
                raise ValueError('Value must be one of: {}'.format(valid_vals))
            self._viewer.set_autocut_params(val)
        else:  # (low, high)
            if len(val) > 2:
                raise ValueError('Value must have length 2.')
            self._viewer.cut_levels(val[0], val[1])

    @property
    def colormap_options(self):
        """List of colormap names."""
        from ginga import cmap
        return cmap.get_names()

    def set_colormap(self, cmap):
        """
        Set colormap to the given colormap name.

        Parameters
        ----------
        cmap : str
            Colormap name. Possible values can be obtained from
            :meth:`colormap_options`.

        """
        self._viewer.set_color_map(cmap)

    @property
    def cursor(self):
        """
        Show or hide cursor information (X, Y, WCS).
        Acceptable values are 'top', 'bottom', or ``None``.
        """
        return self._cursor

    @cursor.setter
    def cursor(self, val):
        if val is None:
            self._jup_coord.layout.visibility = 'hidden'
            self._jup_coord.layout.display = 'none'
        elif val == 'top' or val == 'bottom':
            self._jup_coord.layout.visibility = 'visible'
            self._jup_coord.layout.display = 'flex'
            if val == 'top':
                self.layout.flex_flow = 'column-reverse'
            else:
                self.layout.flex_flow = 'column'
        else:
            raise ValueError('Invalid value {} for cursor.'
                             'Valid values are: '
                             '{}'.format(val, ALLOWED_CURSOR_LOCATIONS))
        self._cursor = val

    @property
    def click_center(self):
        """
        Settable.
        If True, middle-clicking can be used to center.  If False, that
        interaction is disabled.

        In the future this might go from True/False to being a selectable
        button. But not for the first round.
        """
        return self._click_center

    @click_center.setter
    def click_center(self, val):
        if not isinstance(val, bool):
            raise ValueError('Must be True or False')
        elif self.is_marking and val:
            raise ValueError('Cannot set to True while in marking mode')

        if val:
            self.click_drag = False

        self._click_center = val

    # TODO: Awaiting https://github.com/ejeschke/ginga/issues/674
    @property
    def click_drag(self):
        """
        Settable.
        If True, the "click-and-drag" mode is an available interaction for
        panning.  If False, it is not.

        Note that this should be automatically made `False` when selection mode
        is activated.
        """
        return self._click_drag

    @click_drag.setter
    def click_drag(self, value):
        if not isinstance(value, bool):
            raise ValueError('click_drag must be either True or False')
        if self.is_marking:
            raise ValueError('Interactive marking is in progress. Call '
                             'stop_marking() to end marking before setting '
                             'click_drag')
        self._click_drag = value
        bindmap = self._viewer.get_bindmap()
        if value:
            # Only turn off click_center if click_drag is being set to True
            self.click_center = False
            bindmap.map_event(None, (), 'ms_left', 'pan')
        else:
            bindmap.map_event(None, (), 'ms_left', 'cursor')

    @property
    def scroll_pan(self):
        """
        Settable.
        If True, scrolling moves around in the image.  If False, scrolling
        (up/down) *zooms* the image in and out.
        """
        return self._scroll_pan

    @scroll_pan.setter
    def scroll_pan(self, value):
        if not isinstance(value, bool):
            raise ValueError('scroll_pan must be either True or False')

        bindmap = self._viewer.get_bindmap()
        self._scroll_pan = value
        if value:
            bindmap.map_event(None, (), 'pa_pan', 'pan')
        else:
            bindmap.map_event(None, (), 'pa_pan', 'zoom')

    def save(self, filename):
        """
        Save out the current image view to given PNG filename.
        """
        # It turns out the image value is already in PNG format so we just
        # to write that out to a file.
        with open(filename, 'wb') as f:
            f.write(self._jup_img.value)



# Autogenerated by Astropy-affiliated package astrowidgets's setup.py on 2019-10-14 20:41:29 UTC
from __future__ import unicode_literals
import datetime

version = "0.1.1"
githash = "9891e53701142d643276fd3f7cb1b6622fc10bd4"


major = 0
minor = 1
bugfix = 1

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2019, 10, 14, 20, 41, 29)
debug = False

astropy_helpers_version = "3.1.dev1156"



import numpy as np

import pytest

from astropy.io import fits
from astropy.nddata import NDData
from astropy.table import Table

from ginga.ColorDist import ColorDistBase

from ..core import ImageWidget, ALLOWED_CURSOR_LOCATIONS


def test_load_fits():
    image = ImageWidget()
    data = np.random.random([100, 100])
    hdu = fits.PrimaryHDU(data=data)
    image.load_fits(hdu)


def test_load_nddata():
    image = ImageWidget()
    data = np.random.random([100, 100])
    nddata = NDData(data)
    image.load_nddata(nddata)


def test_load_array():
    image = ImageWidget()
    data = np.random.random([100, 100])
    image.load_array(data)


def test_center_on():
    image = ImageWidget()
    x = 10
    y = 10
    image.center_on((x, y))


def test_offset_to():
    image = ImageWidget()
    dx = 10
    dy = 10
    image.offset_to(dx, dy)


def test_zoom_level():
    image = ImageWidget()
    image.zoom_level = 5
    assert image.zoom_level == 5


def test_zoom():
    image = ImageWidget()
    image.zoom_level = 3
    val = 2
    image.zoom(val)
    assert image.zoom_level == 6


@pytest.mark.xfail(reason='Not implemented yet')
def test_select_points():
    image = ImageWidget()
    image.select_points()


def test_get_selection():
    image = ImageWidget()
    marks = image.get_markers()
    assert isinstance(marks, Table) or marks is None


def test_stop_marking():
    image = ImageWidget()
    # This is not much of a test...
    image.stop_marking(clear_markers=True)
    assert image.get_markers() is None
    assert image.is_marking is False


def test_is_marking():
    image = ImageWidget()
    assert image.is_marking in [True, False]
    with pytest.raises(AttributeError):
        image.is_marking = True


def test_start_marking():
    image = ImageWidget()

    # Setting these to check that start_marking affects them.
    image.click_center = True
    assert image.click_center
    image.scroll_pan = False
    assert not image.scroll_pan

    marker_style = {'color': 'yellow', 'radius': 10, 'type': 'cross'}
    image.start_marking(marker_name='something',
                        marker=marker_style)
    assert image.is_marking
    assert image.marker == marker_style
    assert not image.click_center
    assert not image.click_drag

    # scroll_pan better activate when marking otherwise there is
    # no way to pan while interactively marking
    assert image.scroll_pan

    # Make sure that when we stop_marking we get our old
    # controls back.
    image.stop_marking()
    assert image.click_center
    assert not image.scroll_pan

    # Make sure that click_drag is restored as expected
    image.click_drag = True
    image.start_marking()
    assert not image.click_drag
    image.stop_marking()
    assert image.click_drag


def test_add_markers():
    image = ImageWidget()
    table = Table(data=np.random.randint(0, 100, [5, 2]),
                  names=['x', 'y'], dtype=('int', 'int'))
    image.add_markers(table, x_colname='x', y_colname='y',
                      skycoord_colname='coord')


def test_set_markers():
    image = ImageWidget()
    image.marker = {'color': 'yellow', 'radius': 10, 'type': 'cross'}
    assert 'cross' in str(image.marker)
    assert 'yellow' in str(image.marker)
    assert '10' in str(image.marker)


def test_reset_markers():
    image = ImageWidget()
    # First test: this shouldn't raise any errors
    # (it also doesn't *do* anything...)
    image.reset_markers()
    assert image.get_markers() is None
    table = Table(data=np.random.randint(0, 100, [5, 2]),
                  names=['x', 'y'], dtype=('int', 'int'))
    image.add_markers(table, x_colname='x', y_colname='y',
                      skycoord_colname='coord', marker_name='test')
    image.add_markers(table, x_colname='x', y_colname='y',
                      skycoord_colname='coord', marker_name='test2')
    image.reset_markers()
    with pytest.raises(ValueError):
        image.get_markers(marker_name='test')
    with pytest.raises(ValueError):
        image.get_markers(marker_name='test2')


def test_remove_markers():
    image = ImageWidget()
    # Add a tag name...
    image._marktags.add(image._default_mark_tag_name)
    with pytest.raises(ValueError) as e:
        image.remove_markers('arf')
    assert 'arf' in str(e.value)


def test_stretch():
    image = ImageWidget()
    with pytest.raises(ValueError) as e:
        image.stretch = 'not a valid value'
    assert 'must be one of' in str(e.value)

    image.stretch = 'log'
    assert isinstance(image.stretch, (ColorDistBase))


def test_cuts():
    image = ImageWidget()

    # An invalid string should raise an error
    with pytest.raises(ValueError) as e:
        image.cuts = 'not a valid value'
    assert 'must be one of' in str(e.value)

    # Setting cuts to something with incorrect length
    # should raise an error.
    with pytest.raises(ValueError) as e:
        image.cuts = (1, 10, 100)
    assert 'length 2' in str(e.value)

    # These ought to succeed

    image.cuts = 'histogram'
    assert image.cuts == (0.0, 0.0)

    image.cuts = [10, 100]
    assert image.cuts == (10, 100)


def test_colormap():
    image = ImageWidget()
    cmap_desired = 'gray'
    cmap_list = image.colormap_options
    assert len(cmap_list) > 0 and cmap_desired in cmap_list

    image.set_colormap(cmap_desired)


def test_cursor():
    image = ImageWidget()
    assert image.cursor in ALLOWED_CURSOR_LOCATIONS
    with pytest.raises(ValueError):
        image.cursor = 'not a valid option'
    image.cursor = 'bottom'
    assert image.cursor == 'bottom'


def test_click_drag():
    image = ImageWidget()
    # Set this to ensure that click_drag turns it off
    image._click_center = True

    # Make sure that setting click_drag to False does not turn off
    # click_center.

    image.click_drag = False
    assert image.click_center

    image.click_drag = True

    assert not image.click_center

    # If is_marking is true then trying to click_drag
    # should fail.
    image._is_marking = True
    with pytest.raises(ValueError) as e:
        image.click_drag = True
    assert 'Interactive marking' in str(e.value)


def test_click_center():
    image = ImageWidget()
    assert (image.click_center is True) or (image.click_center is False)

    # Set click_drag True and check that click_center affects it appropriately
    image.click_drag = True

    image.click_center = False
    assert image.click_drag

    image.click_center = True
    assert not image.click_drag

    image.start_marking()
    # If marking is in progress then setting click center should fail
    with pytest.raises(ValueError) as e:
        image.click_center = True
    assert 'Cannot set' in str(e.value)

    # setting to False is fine though so no error is expected here
    image.click_center = False


def test_scroll_pan():
    image = ImageWidget()

    # Make sure scroll_pan is actually settable
    for val in [True, False]:
        image.scroll_pan = val
        assert image.scroll_pan is val


def test_save():
    image = ImageWidget()
    filename = 'woot.png'
    image.save(filename)


def test_width_height():
    image = ImageWidget(image_width=250, image_height=100)
    assert image.image_width == 250
    assert image.image_height == 100



from astropy import visualization as aviz
from astropy.nddata.utils import block_reduce
from matplotlib import pyplot as plt

def show_image(image, percl=99, percu=None, figsize=(6, 10), cmap='viridis', log=False):
    """
    Show an image in matplotlib with some basic astronomically-appropriat stretching.
    
    Parameters
    ----------
    image
        The image to show 
    percl : number
        The percentile for the lower edge of the stretch (or both edges if ``percu`` is None)
    percu : number or None
        The percentile for the upper edge of the stretch (or None to use ``percl`` for both)
    figsize : 2-tuple
        The size of the matplotlib figure in inches
    """ 
    if percu is None:
        percu = percl
        percl = 100-percl
        
    if figsize is not None:
        # Rescale the fig size to match the image dimensions, roughly
        image_aspect_ratio = image.shape[0]/image.shape[1]
        figsize = (max(figsize) * image_aspect_ratio, max(figsize))
        
    fig, ax = plt.subplots(1,1, figsize=figsize)
    
    # To preserve details we should *really* downsample correctly and not rely on 
    # matplotlib to do it correctly for us (it won't).
    
    # So, calculate the size of the figure in pixels, block_reduce to roughly that,
    # and display the block reduced image.
    
    # Thanks, https://stackoverflow.com/questions/29702424/how-to-get-matplotlib-figure-size
    fig_size_pix = fig.get_size_inches() * fig.dpi
    
    ratio = (image.shape // fig_size_pix).max()
    
    if ratio < 1:
        ratio = 1
        
    # Divide by the square of the ratio to keep the flux the same in the reduced image
    reduced_data = block_reduce(image, ratio) / ratio**2

    # Of course, now that we have downsampled, the axis limits are changed to match
    # the smaller image size. Setting the extent will do the trick to change the axis display
    # back to showing the actual extent of the image.
    extent = [0, image.shape[1], 0, image.shape[0]]
    if log:
        stretch = aviz.LogStretch()
    else:
        stretch = aviz.LinearStretch()
    norm = aviz.ImageNormalize(reduced_data, interval=aviz.AsymmetricPercentileInterval(percl, percu), 
                                      stretch=stretch)

    plt.colorbar(ax.imshow(reduced_data, norm=norm, origin='lower', cmap=cmap, extent=extent))


import re

import pytest
import numpy as np
from astropy.table import Table, vstack
from astropy.wcs import WCS
from astropy.nddata import CCDData
from astropy.coordinates import SkyCoord

from ..core import ImageWidget, RESERVED_MARKER_SET_NAMES


def _make_fake_ccd(with_wcs=True):
    """
    Generate a CCDData object for use with ImageWidget tests.

    Parameters
    ----------

    with_wcs : bool, optional
        If ``True`` the image will have a WCS attached to it,
        which is useful for some of the marker tests.

    Returns
    -------

    `astropy.nddata.CCDData`
        CCD image
    """
    npix_side = 100
    fake_image = np.random.randn(npix_side, npix_side)
    if with_wcs:
        wcs = WCS(naxis=2)
        wcs.wcs.crpix = (fake_image.shape[0] / 2, fake_image.shape[1] / 2)
        wcs.wcs.ctype = ('RA---TAN', 'DEC--TAN')
        wcs.wcs.crval = (314.275419158, 31.6662781301)
        wcs.wcs.pc = [[0.000153051015113, -3.20700931602e-05],
                      [3.20704370872e-05, 0.000153072382405]]
    else:
        wcs = None

    return CCDData(data=fake_image, wcs=wcs, unit='adu')


def test_setting_image_width_height():
    image = ImageWidget()
    width = 200
    height = 300
    image.image_width = width
    image.image_height = height
    assert image._viewer.get_window_size() == (width, height)


def test_add_marker_does_not_modify_input_table():
    # Regression test for #45
    # Adding markers should not modify the input data table
    image = ImageWidget(image_width=300, image_height=300,
                        pixel_coords_offset=5)
    data = np.random.random([300, 300])
    image.load_array(data)
    x = [20, 30, 40]
    y = [40, 80, 100]
    # Create two separate tables for comparison after add_markers.
    orig_table = Table(data=[x, y], names=['x', 'y'])
    in_table = Table(data=[x, y], names=['x', 'y'])
    image.add_markers(in_table)
    assert (in_table == orig_table).all()


def test_adding_markers_as_world_recovers_with_get_markers():
    """
    Make sure that our internal conversion from world to pixel
    coordinates doesn't mess anything up.
    """
    fake_ccd = _make_fake_ccd(with_wcs=True)
    npix_side = fake_ccd.shape[0]
    wcs = fake_ccd.wcs
    iw = ImageWidget(pixel_coords_offset=0)
    iw.load_nddata(fake_ccd)
    # Get me 100 positions please, not right at the edge
    marker_locs = np.random.randint(10,
                                    high=npix_side - 10,
                                    size=(100, 2))
    marks_pix = Table(data=marker_locs, names=['x', 'y'])
    marks_world = wcs.all_pix2world(marker_locs, 0)
    marks_coords = SkyCoord(marks_world, unit='degree')
    mark_coord_table = Table(data=[marks_coords], names=['coord'])
    iw.add_markers(mark_coord_table, use_skycoord=True)
    result = iw.get_markers()
    # Check the x, y positions as long as we are testing things...
    np.testing.assert_allclose(result['x'], marks_pix['x'])
    np.testing.assert_allclose(result['y'], marks_pix['y'])
    np.testing.assert_allclose(result['coord'].ra.deg,
                               mark_coord_table['coord'].ra.deg)
    np.testing.assert_allclose(result['coord'].dec.deg,
                               mark_coord_table['coord'].dec.deg)


def test_can_set_pixel_offset_at_object_level():
    # The pixel offset below is nonsensical. It is chosen simply
    # to make it easy to check for.
    offset = 3
    image = ImageWidget(image_width=300, image_height=300,
                        pixel_coords_offset=offset)
    assert image._pixel_offset == offset


def test_move_callback_includes_offset():
    # The pixel offset below is nonsensical. It is chosen simply
    # to make it easy to check for.
    offset = 3
    image = ImageWidget(image_width=300, image_height=300,
                        pixel_coords_offset=offset)
    data = np.random.random([300, 300])
    image.load_array(data)
    # Send a fake move to the callback. What gets put in the cursor
    # value should be the event we sent in plus the offset.
    image.click_center = True
    data_x = 100
    data_y = 200
    image._mouse_move_cb(image._viewer, None, data_x, data_y)
    output_contents = image._jup_coord.value
    x_out = re.search(r'X: ([\d\.\d]+)', output_contents)
    x_out = x_out.groups(1)[0]
    y_out = re.search(r'Y: ([\d\.\d]+)', output_contents)
    y_out = y_out.groups(1)[0]
    assert float(x_out) == data_x + offset
    assert float(y_out) == data_y + offset


def test_can_add_markers_with_names():
    """
    Test a few things related to naming marker sets
    """
    npix_side = 200
    image = ImageWidget(image_width=npix_side,
                        image_height=npix_side)
    x = np.array([20, 30, 40])
    y = np.array([40, 80, 100])

    # This should succeed without error
    image.add_markers(Table(data=[x, y], names=['x', 'y']),
                      marker_name='nonsense')

    # The name 'nonsense', and nothing else, should be in the
    # set of markers.
    assert set(['nonsense']) == image._marktags

    # Add more markers with the same name
    # This should succeed without error
    image.add_markers(Table(data=[x, y], names=['x', 'y']),
                      marker_name='nonsense')

    # check that we get the right number of markers
    marks = image.get_markers(marker_name='nonsense')
    assert len(marks) == 6

    # Make sure setting didn't change the default name
    assert image._default_mark_tag_name == 'default-marker-name'

    # Try adding markers without a name
    image.add_markers(Table(data=[x, y], names=['x', 'y']))
    assert image._marktags == set(['nonsense', image._default_mark_tag_name])

    # Delete just the nonsense markers
    image.remove_markers('nonsense')

    assert 'nonsense' not in image._marktags
    assert image._default_mark_tag_name in image._marktags

    # Add the nonsense markers back...
    image.add_markers(Table(data=[x, y], names=['x', 'y']),
                      marker_name='nonsense')
    # ...and now delete all of the markers
    image.reset_markers()
    # We should have no markers on the image
    assert image._marktags == set()

    # Simulate a mouse click and make sure the expected marker
    # name has been added.
    data_x = 50
    data_y = 50
    image._is_marking = True
    image._mouse_click_cb(image._viewer, None, data_x, data_y)
    assert image._interactive_marker_set_name in image._marktags


def test_mark_with_reserved_name_raises_error():
    npix_side = 200
    image = ImageWidget(image_width=npix_side,
                        image_height=npix_side)
    x = np.array([20, 30, 40])
    y = np.array([40, 80, 100])
    for name in RESERVED_MARKER_SET_NAMES:
        with pytest.raises(ValueError):
            image.add_markers(Table(data=[x, y], names=['x', 'y']),
                              marker_name=name)


def test_get_marker_with_names():
    # Check a few ways of getting markers out
    npix_side = 200
    image = ImageWidget(image_width=npix_side,
                        image_height=npix_side)

    x = np.array([20, 30, 40])
    y = np.array([40, 80, 100])
    input_markers = Table(data=[x, y], names=['x', 'y'])
    # Add some markers with our own name
    image.add_markers(input_markers, marker_name='nonsense')

    # Add same markers without a name so that name defaults to
    # image._default_mark_tag_name
    image.add_markers(input_markers)

    # Add pseudo-interactive points
    image._is_marking = True
    for data_x, data_y in input_markers:
        image._mouse_click_cb(image._viewer, None, data_x, data_y)

    # Should have three sets of markers: nonsense, default non-interactive,
    # interactive
    assert len(image._marktags) == 3

    for marker in image._marktags:
        out_table = image.get_markers(marker_name=marker)
        # No guarantee markers will come back in the same order, so sort them.
        out_table.sort('x')
        assert (out_table['x'] == input_markers['x']).all()
        assert (out_table['y'] == input_markers['y']).all()

    # Get all of markers at once
    all_marks = image.get_markers(marker_name='all')

    # That should have given us three copies of the input table
    expected = vstack([input_markers] * 3, join_type='exact')

    # Sort before comparing
    expected.sort(['x', 'y'])
    all_marks.sort(['x', 'y'])

    assert (expected['x'] == all_marks['x']).all()
    assert (expected['y'] == all_marks['y']).all()


def test_unknown_marker_name_error():
    """
    Regression test for https://github.com/astropy/astrowidgets/issues/97

    This particular test checks that getting a marker name that
    does not exist raises an error.
    """
    iw = ImageWidget()
    bad_name = 'not a real marker name'
    with pytest.raises(ValueError) as e:
        iw.get_markers(marker_name=bad_name)

    assert f"No markers named '{bad_name}'" in str(e.value)


def test_marker_name_has_no_marks_warning():
    """
    Regression test for https://github.com/astropy/astrowidgets/issues/97

    This particular test checks that getting an empty table gives a
    useful warning message.
    """
    iw = ImageWidget()
    bad_name = 'empty marker set'
    iw.start_marking(marker_name=bad_name)

    with pytest.warns(UserWarning) as record:
        iw.get_markers(marker_name=bad_name)

    assert f"Marker set named '{bad_name}' is empty" in str(record[0].message)


def test_empty_marker_name_works_with_all():
    """
    Regression test for https://github.com/astropy/astrowidgets/issues/97

    This particular test checks that an empty table doesn't break
    marker_name='all'. The bug only comes up if there is a coordinate
    column, so use a fake image a WCS.
    """
    iw = ImageWidget()
    fake_ccd = _make_fake_ccd(with_wcs=True)
    iw.load_nddata(fake_ccd)

    x = np.array([20, 30, 40])
    y = np.array([40, 80, 100])
    input_markers = Table(data=[x, y], names=['x', 'y'])
    # Add some markers with our own name
    iw.add_markers(input_markers, marker_name='nonsense')

    # Start marking to create a new marker set that is empty
    iw.start_marking(marker_name='empty')

    marks = iw.get_markers(marker_name='all')
    assert len(marks) == len(x)
    assert 'empty' not in marks['marker name']


def test_add_single_marker():
    """
    Test a few things related to naming marker sets
    """
    fake_ccd = _make_fake_ccd(with_wcs=True)
    npix_side = fake_ccd.shape[0]
    wcs = fake_ccd.wcs
    iw = ImageWidget(pixel_coords_offset=0)
    iw.load_nddata(fake_ccd)
    # Get me 100 positions please, not right at the edge
    marker_locs = np.random.randint(10,
                                    high=npix_side - 10,
                                    size=(100, 2))
    marks_world = wcs.all_pix2world(marker_locs, 0)
    marks_coords = SkyCoord(marks_world, unit='degree')
    mark_coord_table = Table(data=[marks_coords], names=['coord'])
    iw.add_markers(mark_coord_table[0], use_skycoord=True)
